import collections

import kivy.app
from kivy.clock import Clock
import kvaut.automator.factory as factory
import kvaut.automator.tapper as tapper


class WidgetAutomator(object):

    def __init__(self, target):
        self._target = target

    def global_center(self):
        return self._target.to_window(self._target.center_x, self._target.center_y)

    @property
    def values(self):
        return [self._target.id]

    def is_match(self, value=None):
        if value in self.values:
            return True

        return False

    def get_children(self):
        return [factory.automate(c) for c in self._target.children];
		
    def tap(self):
        tapper.tap(self)

    def to_json(self, is_recursive=False):
        json=collections.OrderedDict([ 
            ('type',self._target.__class__.__name__), 
            ('values',self.values),  
        ])

        if is_recursive:
            children_json = []
            for child in self.get_children():
                children_json.append(child.to_json(is_recursive))
            json['children'] = children_json

        return json




import collections

import kvaut.automator.factory as factory
import kvaut.automator.widget


class LabelAutomator(kvaut.automator.widget.WidgetAutomator):

    def __init__(self, target):
        self._target = target

    @property
    def values(self):
        return [self._target.text, self._target.id]




from __future__ import unicode_literals
import importlib
import logging


logger = logging.getLogger(__name__)

def automate(target):
    automators = ['Label','Widget']

    for class_name in automators:
        try:
            kivy_module = importlib.import_module('kivy.uix.{}'.format(class_name.lower()))
            kivy_type = getattr(kivy_module, class_name)

            automator_module = importlib.import_module('kvaut.automator.{}'.format(class_name.lower()))
            automator_type = getattr(automator_module, '{}Automator'.format(class_name))

            if kivy_type is not None and automator_type is not None and isinstance(target, kivy_type):
                return automator_type(target)
        except Exception as ex:
            logger.error('could not import {}: {}'.format(class_name, ex))
            continue

    raise Exception('No automator found for {}'.format(target))




from __future__ import unicode_literals
import os
import threading
import logging

import bottle

import kvaut.automator.factory as factory


@bottle.get("/ping")
def ping():
    if get_root_widget() is None:
        bottle.abort(503, "Still booting up, try again later")

    return 'Ping!'

@bottle.get("/tree")
def tree():
    widget = get_root_widget()
    if widget is not None:
        return widget.to_json(is_recursive=True)

@bottle.post("/find_element")
def find_element():
    found_json = {}
    value = get_query_value('value')

    widget = find_widget_in(get_root_widget(), value=value)
    if widget is not None:
        found_json = widget.to_json()

    logger.debug('found element: {}'.format(found_json))
    return found_json

@bottle.post("/tap")
def tap():
    found_json = {}
    value = get_query_value('value')

    widget = find_widget_in(get_root_widget(), value=value)
    if widget is not None:
        found_json = widget.to_json()
        widget.tap()

    return found_json

def get_query_value(name):
    if bottle.request.json is not None:
        return bottle.request.json.get('query', {}).get(name, '')

    return ''

def get_root_widget():
    import kivy.app
    app = kivy.app.App.get_running_app()
    return factory.automate(app.root)

def find_widget_in(parent, value=None):
    if parent is None:
        return None

    if parent.is_match(value=value):
        return parent

    for child in parent.get_children():
        if child.is_match(value=value):
            return child

        found_widget = find_widget_in(child, value=value)
        if found_widget is not None:
            return found_widget

    return None

def start_automation_server():
    debug = log_level_name is 'DEBUG'
    thread = threading.Thread(target=bottle.run, kwargs={'host':'0.0.0.0', 'port':5155, 'quiet':(not debug), 'debug':debug})
    thread.setDaemon(True)
    thread.start()

log_levels = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
}
log_level_name = os.environ.get('KVAUT_LOG', 'ERROR')
logging.basicConfig(level=log_levels[log_level_name])
logger = logging.getLogger(__name__)
logger.setLevel(log_levels[log_level_name])
logger.info('kvaut log level is {}'.format(log_level_name))



from kivy.input.motionevent import MotionEvent

class FakeMotionEvent(MotionEvent):
 
    def depack(self, args):
        self.is_touch = True
        self.sx = args['x']
        self.sy = args['y']
        self.profile = ['pos']
        super(FakeMotionEvent, self).depack(args)



import time


def wait_for(condition, retries=6, loop_delay=0.5):
    met_condition = False
    for i in range(0, retries):
        time.sleep(loop_delay)
        met_condition = condition()
        if met_condition:
            break

    return met_condition



import sys
try:
    import ez_setup
    ez_setup.use_setuptools()
except ImportError:
    pass

from setuptools import setup

setup(
    name='kvaut',
    version='0.0.6',
    author='Gary Johnson',
    author_email = 'gary@gjtt.com',
    description = 'Automation for BDD testing Kivy apps',
    install_requires=['bottle'],
    tests_require=['bottle'],
    license = 'MIT License',
    packages = ['kvaut', 'kvaut.automator', 'kvaut.errors', 'kvaut.helpers'],
    )



from __future__ import unicode_literals
import os
import time
import json
import logging

import requests

from kvaut.helpers.wait import *
import kvaut.errors


JSON_HEADERS = {'Content-type': 'application/json', 'Accept': 'application/json' }
SERVER="http://0.0.0.0:5155"
TIMEOUT=3

def assert_is_visible(target):
    if not wait_for(lambda: find_element(target)):
        raise kvaut.errors.AssertionError('Could not find element matching \"{}\"'.format(target))

def find_element(target):
    found_element = None

    try:
        body = {'query':{'value':target }}
        response = requests.post(url_for('/find_element'), data=json.dumps(body), headers=JSON_HEADERS, timeout=TIMEOUT)
        found_element = response.json()
        if len(found_element.keys()) == 0:
            found_element = None
    except Exception as ex:
        logger.debug('error occurred attempting to find element matching \"{}\": {}'.format(target, ex))

    print("found element {}".format(found_element))
    return found_element

def tap(target):
    assert_is_visible(target)

    tapped_element = None
    try:
        body = {'query':{'value':target }}
        response = requests.post(url_for('/tap'), data=json.dumps(body), headers=JSON_HEADERS, timeout=TIMEOUT)
        tapped_element = response.json()
        if len(tapped_element.keys()) == 0:
            tapped_element = None
    except Exception as ex:
        logger.debug('error occurred attempting to tap element matching \"{}\": {}'.format(target, ex))

    print("tapped element {}".format(tapped_element))
    return tapped_element

def wait_for_automation_server():
    logger.info('waiting for automation server')

    def automation_server_exists():
        try:
            response = requests.get(url_for("/ping"), timeout=TIMEOUT)
            return response.status_code == 200
        except Exception as ex:
            logger.debug('error occurred while waiting for automation server: {}'.format(ex))
            return False

    if not wait_for(automation_server_exists):
        raise kvaut.errors.ServerNotFoundError('kvaut timed out waiting for automation server')

def url_for(path):
    return "{}{}".format(SERVER, path)

log_levels = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
}
log_level_name = os.environ.get('KVAUT_LOG', 'ERROR')
logging.basicConfig(level=log_levels[log_level_name])
logger = logging.getLogger(__name__)
logger.setLevel(log_levels[log_level_name])
logger.info('kvaut log level is {}'.format(log_level_name))



from functools import partial

from kivy.app import App
from kivy.clock import Clock, mainthread
from kivy.base import EventLoop

from kvaut.automator.fake_motion_event import FakeMotionEvent


def tap(automator):
    Clock.schedule_once(partial(_tap_on_ui_thread, automator), 0)

@mainthread
def _tap_on_ui_thread(automator, args):
    global_x,global_y = automator.global_center()
    app = App.get_running_app()
    relative_pos = {"x": global_x / app.root.width, "y" : global_y / app.root.height}

    touch = FakeMotionEvent("fake", 1, relative_pos)
     
    EventLoop.post_dispatch_input("begin", touch)
    EventLoop.post_dispatch_input("end", touch)
